(*
#$Id: datatypes.sml 34 2011-07-21 17:52:28Z ns441 $
#
#File containing supporting definitions for the "prefactor" experiment.
#
#Nik Sultana
#www.cl.cam.ac.uk/~ns441
#
*)

signature PCFP1_SYNTAX = 
sig
  (*filling contains whitespace, parentheses, comments.*)
  type filling  = string list

  (*indexes a character is a program (seen as a list of characters)*)
  type index = int;
  (*describes an interval between a start and end index in a program*)
  type range = int;
  (*position*)
  type pos = index * range

  datatype pre_typ = 
     T_Nat              (*nat*)
   | T_Fun of typ * typ (*T1 -> T2*)
   | T_Unit             (*unit*)
   | T_Sum of typ * typ (*T1 + T2*)
  and typ = Type of pre_typ * filling

  datatype ''a pre_trm = 
      E_Var of ''a                                             (*x*)
    | E_Lam of ''a * typ * ''a trm                             (*fn x : T => E*)
    | E_App of ''a trm * ''a trm                               (*E1 E2*)
    | E_Un                                                     (*unity*)
    | E_Fix of ''a * typ * ''a trm                             (*fix x : T => E*)
    | E_InL of typ * ''a trm                                   (*left T E*)
    | E_InR of typ * ''a trm                                   (*right T E*)
    | E_Case of (''a trm) * ''a * (''a trm) * ''a * (''a trm)  (*case [E1 <= y] E3 [z => E2]*)
    | E_Zero                                                   (*zero*)
    | E_Succ of ''a trm                                        (*succ E*)
    | E_Pred of ''a trm                                        (*pred E*)
    | E_IfZero of (''a trm) * (''a trm) * (''a trm)            (*ifZero E1 then E2 else E3*)
  and ''a trm = Term of ''a pre_trm * filling

  type term = string trm

  val print_typ : typ -> string
  val print_trm : term -> string

  (*describes the path down a tree in terms of which branches to take*)
  type path = int list

  (*attempts to extract a subprogram indicated by a range of characters in the program text*)
  val point : term -> pos -> (term * path) option

  val debugging : bool ref
  val debug_print : string -> unit

  datatype expression = TermExp of term
		      | TypeExp of typ

  val term_of : expression -> term
  val type_of : expression -> typ

  val rebracket : typ -> typ
  val reformat : term -> term

  exception EMPTY_PATH
  exception WAYWARD_PATH
  val insert : term -> path -> term -> term

  (*get and add surrounding whitespace of terms*)
  val get_surr_ws : term -> string * string
  val add_surr_ws : term -> string * string -> term

  val t_nat_: string
  val t_unit_: string
  val t_sum_: string
  val t_fn_: string
end


structure PCFP1_Syntax : PCFP1_SYNTAX = 
struct

val debugging = ref false
fun debug_print str = 
  if !debugging then 
    print ("::" ^ str)
	else
	  ()

val typechecking = ref true

type filling  = string list

datatype pre_typ = 
    T_Nat              (*nat*)
  | T_Fun of typ * typ (*T1 -> T2*)
  | T_Unit             (*unit*)
  | T_Sum of typ * typ (*T1 + T2*)
and typ = Type of pre_typ * filling

datatype ''a pre_trm =
    E_Var of ''a                  (*x*)
  | E_Lam of ''a * typ * ''a trm  (*fn x : T => E*)
  | E_App of ''a trm * ''a trm    (*E1 E2*)
  | E_Un                          (*unity*)
  | E_Fix of ''a * typ * ''a trm  (*fix x : T => E*)
  | E_InL of typ * ''a trm        (*left T E*)
  | E_InR of typ * ''a trm        (*right T E*)
  | E_Case of (''a trm) * ''a * (''a trm) * ''a * (''a trm) (*case (E1 <= y) E3 (z => E2)*)
  | E_Zero                        (*zero*)
  | E_Succ of ''a trm             (*succ E*)
  | E_Pred of ''a trm             (*pred E*)
  | E_IfZero of (''a trm) * (''a trm) * (''a trm) (*ifZero E1 then E2 else E3*)
and ''a trm = Term of ''a pre_trm * filling

(*"tokens"*)
val t_nat_      =  "nat"
val t_unit_     =  "unit"
val t_sum_      =  "+"
val t_fn_       =  "->"

val e_fn_       =  "fn"
val e_colon_    =  ":"
val e_rarrow_   =  "=>"
val e_unity_    =  "unity"
val e_fix_      =  "fix"
val e_zero_     =  "zero"
val e_succ_     =  "succ"
val e_pred_     =  "pred"
val e_ifZero_   =  "ifZero"
val e_then_     =  "then"
val e_else_     =  "else"
val e_left_     =  "left"
val e_right_    =  "right"
val e_case_     =  "case"
val e_lbrack_   =  "["
val e_rbrack_   =  "]"
val e_larrow_   =  "<="

local
  fun get_next_fill filling () = 
    if List.null (!filling) then
      ""
    else
      let
        val head = hd (!filling)
        val _    = filling := tl (!filling)
      in head end
in
  fun print_typ (Type ty) =
    let
      val filling = ref []
      val next_fill = get_next_fill filling 
    in
      case ty of
          (T_Nat, fill) => 
            (filling := fill
            ;next_fill () ^ t_nat_ ^ next_fill ())
        | (T_Unit, fill) => 
            (filling := fill
            ;next_fill () ^ t_unit_ ^ next_fill ())
        | (T_Fun (ty1,ty2), fill) => 
            (filling := fill
            ;next_fill () ^ print_typ ty1 ^ t_fn_ ^ print_typ ty2 ^ next_fill ())
        | (T_Sum (ty1,ty2), fill) => 
            (filling := fill
            ;next_fill () ^ print_typ ty1 ^ t_sum_ ^ print_typ ty2 ^ next_fill ())
    end
  
  fun print_trm (Term term) =
    let
      val filling = ref []
      val next_fill = get_next_fill filling 
    in
      case term of
          (E_Var v, fill) => 
            (filling := fill; 
            next_fill () ^ v ^ next_fill ()
            )
        | (E_Lam (v,ty,t), fill) => 
            (filling := fill; 
             next_fill () ^ e_fn_ ^ next_fill () ^ v ^ 
             next_fill () ^ e_colon_ ^ next_fill () ^ 
             print_typ ty ^ next_fill () ^ e_rarrow_ ^ 
             next_fill () ^ print_trm t ^ next_fill ()
            )
        | (E_App (t1,t2), fill) => 
            (filling := fill; 
             next_fill () ^ print_trm t1 ^ next_fill () ^ 
             print_trm t2 ^ next_fill ()
            )
        | (E_Fix (v,ty,t), fill) => 
            (filling := fill; 
             next_fill () ^ e_fix_ ^ next_fill () ^ v ^
             next_fill () ^ e_colon_ ^ next_fill () ^
             print_typ ty ^ next_fill () ^ e_rarrow_ ^
             next_fill () ^ print_trm t ^ next_fill ()
            )
        | (E_Un, fill) => 
            (filling := fill; 
             next_fill () ^ e_unity_ ^ next_fill ()
            )
        | (E_Zero, fill) => 
            (filling := fill; 
             next_fill () ^ e_zero_ ^ next_fill ()
            )
        | (E_Succ t, fill) => 
            (filling := fill; 
             next_fill () ^ e_succ_ ^ next_fill () ^ 
             print_trm t ^ next_fill ()
            )
        | (E_Pred t, fill) => 
            (filling := fill; 
             next_fill () ^ e_pred_ ^ next_fill () ^ 
             print_trm t ^ next_fill ()
            )
        | (E_IfZero (t1,t2,t3), fill) => 
            (filling := fill; 
             next_fill () ^ e_ifZero_ ^ next_fill () ^
             print_trm t1 ^ next_fill () ^ e_then_ ^ 
             next_fill () ^ print_trm t2 ^ next_fill () ^ 
             e_else_ ^ next_fill () ^ print_trm t3 ^ next_fill ()
            )
        | (E_InL (ty,t), fill) => 
            (filling := fill; 
             next_fill () ^ e_left_ ^ next_fill () ^ print_typ ty ^
             next_fill () ^ print_trm t ^ next_fill ()
            )
        | (E_InR (ty,t), fill) => 
            (filling := fill; 
             next_fill () ^ e_right_ ^ next_fill () ^ print_typ ty ^
             next_fill () ^ print_trm t ^ next_fill ()
            )
        | (E_Case (tL,vL,t,vR,tR), fill) => 
            (filling := fill; 
             next_fill () ^ e_case_ ^ next_fill () ^ e_lbrack_ ^ next_fill () ^ 
             print_trm tL ^ next_fill () ^ e_larrow_ ^ next_fill () ^ vL ^ 
             next_fill () ^ e_rbrack_ ^  next_fill () ^ print_trm t ^ 
             next_fill () ^ e_lbrack_ ^ next_fill () ^ vR ^ next_fill () ^
             e_rarrow_ ^ next_fill () ^ print_trm tR ^ next_fill () ^ e_rbrack_ ^ 
             next_fill ()
            )
    end
end

(** Adaptation with extracted code **)

fun leftFill fill = (*extract leftmost filling*)
  if List.null fill then
    ""
  else
    hd fill

fun rightFill fill = (*extract rightmost filling*)
  if List.null fill then
    ""
  else
    List.last fill

type path = int list

type index = int
type range = int
type pos = index * range

(*sum the sizes of strings in a list*)
fun sizes list = foldr (op +) 0 (map String.size list)

(*size of the syntax of a particular type expression*)
fun synsize_type (Type (ty, fill)) = 
  let
    val symbol_size = 
      case ty of
         T_Nat        => String.size t_nat_
       | T_Fun (_, _) => String.size t_fn_
       | T_Unit       => String.size t_unit_
       | T_Sum (_, _) => String.size t_sum_
    val children_size = 
      case ty of
         T_Nat            => 0
       | T_Fun (ty1, ty2) => synsize_type ty1 + synsize_type ty2
       | T_Unit           => 0
       | T_Sum (ty1, ty2) => synsize_type ty1 + synsize_type ty2
  in sizes fill + symbol_size + children_size
  end
    
(*size of the syntax of a particular term expression*)
fun synsize_term (Term (t,fill)) = 
  let
    val symbol_size = 
      case t of
         E_Var v              => String.size v
       | E_Lam (v,_,_)        => sizes [e_fn_, v, e_colon_, e_rarrow_]
       | E_App (_,_)          => 0
       | E_Un                 => String.size e_unity_
       | E_Fix (v,_,_)        => sizes [e_fix_, v, e_colon_, e_rarrow_]
       | E_InL (_,_)          => String.size e_left_
       | E_InR (_,_)          => String.size e_right_
       | E_Case (_,vL,_,vR,_) => sizes [e_case_, e_lbrack_, e_larrow_, vL,
                                 e_rbrack_, e_lbrack_, vR, e_rarrow_, e_rbrack_]
       | E_Zero               => String.size e_zero_
       | E_Succ _             => String.size e_succ_
       | E_Pred _             => String.size e_pred_
       | E_IfZero (_,_,_)     => sizes [e_ifZero_, e_then_, e_else_]                                      
    val children_size = 
      case t of
        E_Var _                 => 0
      | E_Lam (_,ty,t')         => synsize_type ty + synsize_term t'
      | E_App (t1,t2)           => synsize_term t1 + synsize_term t2
      | E_Un                    => 0
      | E_Fix (_,ty,t')         => synsize_type ty + synsize_term t'
      | E_InL (ty,t')           => synsize_type ty + synsize_term t'
      | E_InR (ty,t')           => synsize_type ty + synsize_term t'
      | E_Case (tL,_,t',_,tR)   => synsize_term tL + synsize_term t' + 
                                   synsize_term tR
      | E_Zero                  => 0
      | E_Succ t'               => synsize_term t'
      | E_Pred t'               => synsize_term t'
      | E_IfZero (t1,t2,t3)     => synsize_term t1 + synsize_term t2 + 
                                    synsize_term t3
  in sizes fill + symbol_size + children_size
  end

type term = string trm

fun leftBracketBuffer fill = (*counts whitespace until first "("*)
  let 
    val fill' = String.explode fill
    fun findLeftBracket (c::rest) i = 
          if c = #"(" then 
            i
          else 
            findLeftBracket rest (i+1)
      | findLeftBracket _ i = i (*raise List.Empty*)
  in 
    if List.null fill' then 
      0 
    else 
      findLeftBracket fill' 1
  end
    
fun rightBracketBuffer fill = (*counts whitespace following last ")"*)
  let 
    val fill' = List.rev (String.explode fill)
    fun findRightBracket (c::rest) i = 
	        if c = #")" then 
	          i
	        else 
	          findRightBracket rest (i+1)
      | findRightBracket _ _ = 0 (*raise List.Empty*)
    in 
      if List.null fill' then 
        0 
      else 
        List.length fill' - findRightBracket fill' 1
  end

(*aggregate size of fillings except for the last "n" ones; "last" since we're working from L to R*)
fun aggrSizeExceptForLastn _ [] = 0 
  | aggrSizeExceptForLastn n fill = 
      let
        val sizeDiff = List.length fill - n
        fun fill' () = List.take (fill, sizeDiff)
      in 
        if n <= 0 then raise 
          List.Empty
        else 
          sizes (fill' ())
      end
    
fun termheadOf t = 
  case t of
     E_Var _  => "E_Var"
   | E_Lam _  => "E_Lam"
   | E_App _  => "E_App"
   | E_Un     => "E_Un"
   | E_Fix _  => "E_Fix"
   | E_InL _  => "E_InL"
   | E_InR _  => "E_InR"
   | E_Case _ => "E_Case"
   | E_Zero   => "E_Zero"
   | E_Succ _ => "E_Succ"
   | E_Pred _ => "E_Pred"
   | E_IfZero _ => "E_IfZero"
                     
fun point' (t as (Term (t',fill))) (index,range) path = 
  let
    fun postIndex t offset = (*find if the end of a term exceeds the index'*)
      let
        val size = synsize_term t + offset
        val outcome = size < index
        val _ = 
          debug_print ("postIndex consulted; decided " ^ Bool.toString outcome ^ 
             " since size of subterm considered is "^ Int.toString size ^"\n")
      in outcome
      end
    (* switching to using this would require tighter "highlighting" around an expression
    val leftpad = ~(leftBracketBuffer (leftFill fill))
    val rightpad = rightBracketBuffer (rightFill fill)
     *)
    val leftpad = ~(String.size (leftFill fill)) (*size of leftmost filling*)
    val rightpad = String.size (rightFill fill)  (*..and of rightmost filling*)

    val cover_max = synsize_term t (*full size of this term (including all filling)*)
    val cover_min = cover_max - (rightpad - leftpad) (*..and excluding the filling at the boundaries*)

    val s = "step " ^ Int.toString(List.length path) ^ ": looking at " ^
      print_trm t ^ "\nhead is " ^ termheadOf t' ^ "\nindex=" ^
       Int.toString index ^ " range=" ^ Int.toString range ^ 
       "; lp=" ^ Int.toString leftpad ^ " rp=" ^ Int.toString rightpad ^
       "; ci=" ^ Int.toString cover_min ^ " ca=" ^ Int.toString cover_max ^
       " path=" ^ String.concat (map Int.toString path) ^ "\n"
              
    val _ = debug_print s
    fun failMsg () = debug_print "'point' failed\n"
                                                                          
    (*
      go through subterms, making sure to update the index correctly (this needs
      to be done on a per-expression basis and works according to "printed" 
      syntax (i.e. takes into account the fillings))
    *)
    fun checkChildren () = 
      case t' of
         E_Var _ => NONE (*since a variable doesnt have subterms*)
       | E_Lam (v, ty, t') => (*note that type-based "point'" will be implemented separately*)
          let 
            val offset = String.size v (*the only expression subterm is t', so calculate the syntax offset until it starts in order to modify index accordingly*)
			        + synsize_type ty
			        + sizes [e_fn_, e_colon_, e_rarrow_]
			        + aggrSizeExceptForLastn 1 fill
			    in point' t' (index - offset, range) (3 :: path)
			    end
       | E_App (t1,t2) => (*this is where range comes in particularly useful since it helps disambiguate what the user is pointing at*)
          if (postIndex t1 leftpad) then (*if index is beyond the total size of t1 then look at t2, otherwise look at t1*)
            let 
              val offset = synsize_term t1
			          + aggrSizeExceptForLastn 1 fill
			      in point' t2 (index - offset, range) (2 :: path)
			      end
			    else point' t1 (index + leftpad, range) (1 :: path)
       | E_Un => NONE
       | E_Fix (v, ty, t') => 
          let 
           val offset = String.size v
            + synsize_type ty
            + sizes [e_fix_, e_colon_, e_rarrow_]
            + aggrSizeExceptForLastn 1 fill
          in
            point' t' (index - offset, range) (3 :: path)
          end
       | E_InL (ty, t') =>
          let
             val offset = sizes [e_left_]
               + synsize_type ty
               + aggrSizeExceptForLastn 1 fill
          in point' t' (index - offset, range) (2 :: path)
          end
       | E_InR (ty, t') =>
          let 
            val offset = sizes [e_right_]
              + synsize_type ty
              + aggrSizeExceptForLastn 1 fill
          in point' t' (index - offset, range) (2 :: path)
          end
       | E_Case (tL, vL, t' ,vR ,tR ) => 
          if (postIndex tL 0) then 
            ( 
              if (postIndex t' 0) then 
                let 
                   val offset = sizes [e_case_, e_lbrack_, e_larrow_, e_rbrack_,
                      e_lbrack_, e_rarrow_]
                     + synsize_term tL
                     + synsize_term t'
                     + String.size vL
                     + String.size vR
                     + aggrSizeExceptForLastn 2 fill
                in point' tR (index-offset,range) (3::path)
                end
              else 
                let 
                  val offset = sizes [e_case_,e_lbrack_,e_larrow_,e_rbrack_]
                    + synsize_term tL
                    + String.size vL
                    + aggrSizeExceptForLastn 6 fill
                  in point' t' (index-offset,range) (2::path)
                end
            )
          else 
            let 
              val offset = sizes [e_case_,e_lbrack_]
                + aggrSizeExceptForLastn 10 fill
            in point' tL (index-offset,range) (1::path)
            end
       | E_Zero                   => NONE
       | E_Succ t'                => 
          let 
            val offset = sizes [e_succ_]
              + aggrSizeExceptForLastn 1 fill
          in point' t' (index-offset,range) (1::path)
          end
       | E_Pred t'                => 
          let 
            val offset = sizes [e_pred_]
              + aggrSizeExceptForLastn 1 fill
          in point' t' (index-offset,range) (1::path)
          end
       | E_IfZero (t1,t2,t3)      => 
          if (postIndex t1 0) then 
            ( 
              if (postIndex t2 0) then 
                let 
                  val offset = sizes [e_ifZero_,e_then_,e_else_]
                    + synsize_term t1
                    + synsize_term t2
                    + aggrSizeExceptForLastn 1 fill
                in point' t3 (index-offset,range) (3::path)
                end
              else 
                let 
                  val offset = sizes [e_ifZero_,e_then_]
                    + synsize_term t1
                    + aggrSizeExceptForLastn 3 fill
                in point' t2 (index-offset,range) (2::path)
                end
            )
          else 
            let 
              val offset = sizes [e_ifZero_]
                + aggrSizeExceptForLastn 5 fill
            in point' t1 (index-offset,range) (1::path)
            end
             
    val index' = index + leftpad

  in 
    (*the start of this term takes place from first "(" up to first letter of first keyword*)
    if leftpad <= index' andalso index' <= 1
    then (
      if (cover_min <= range andalso range <= cover_max) then 
        SOME (t, List.rev (0 :: path))
      else 
        (debug_print "range doesn't fall on end; checking subterms\n"
        ; checkChildren ())
      )
    else (
      if (range > cover_max) then 
        (failMsg (); NONE) (*expression can't be a subexpression, the range is too wide!*)
      else (*therefore range < cover, the expression sought must be a subexpression*)
        (debug_print "index doesn't fall on start; checking subterms\n"
        ; checkChildren ())
      )
  end
    
fun point t pos = point' t pos []

datatype expression = TermExp of term | TypeExp of typ

fun term_of (TermExp x) = x
  | term_of _           = raise Option

fun type_of (TypeExp x) = x
  | type_of _           = raise Option

fun preserveBrackets fills = (*remove everything except for brackets*)
    let 
      fun erase fill =
        String.implode (List.filter (fn x => (x = #"(") orelse (x = #")"))
          (String.explode fill))
    in map erase fills
    end

(*remove all filling*)
fun flushFilling_type (Type (ty,fills)) = (Type (ty,[](*preserveBrackets fills*)))

fun rebracket (Type (ty,_)) = 
  let
    val ty'= 
      case ty of 
         T_Nat            => T_Nat
       | T_Fun (ty1,ty2)  => T_Fun (rebracket ty1,rebracket ty2)
       | T_Unit           => T_Unit
       | T_Sum (ty1,ty2)  => T_Sum (rebracket ty1,rebracket ty2)
    val fills'=
      case ty of 
         T_Nat            => []
       | T_Fun (ty1,ty2)  => ["(",")"]
       | T_Unit           => []
       | T_Sum (ty1,ty2)  => ["(",")"]
  in Type (ty', fills') end

fun reformat (Term (t,_)) = (*"prettyprints"*)
  let
    val t'= 
      case t of 
         E_Var v                 => E_Var v
       | E_Lam (v,ty,t')         => E_Lam (v, rebracket ty, reformat t')
       | E_App (t1,t2)           => E_App (reformat t1, reformat t2)
       | E_Un                    => E_Un
       | E_Fix (v,ty,t')         => E_Fix (v, rebracket ty, reformat t')
       | E_InL (ty,t')           => E_InL (rebracket ty, reformat t')
       | E_InR (ty,t')           => E_InR (rebracket ty, reformat t')
       | E_Case (tL,vL,t',vR,tR) => E_Case (reformat tL, vL, reformat t', vR,
                                      reformat tR)
       | E_Zero                  => E_Zero
       | E_Succ t'               => E_Succ (reformat t')
       | E_Pred t'               => E_Pred (reformat t')
       | E_IfZero (t1,t2,t3)     => E_IfZero (reformat t1, reformat t2,
                                      reformat t3)
    val fills'=
      case t of 
         E_Var _  => []
       | E_Lam _  => ["("," "," "," "," "," ",")"]
       | E_App _  => ["("," ",")"]
       | E_Un     => []
       | E_Fix _  => ["("," "," "," "," "," ",")"]
       | E_InL _  => ["("," "," ",")"]
       | E_InR _  => ["("," "," ",")"]
       | E_Case _ => ["("," "," "," "," "," "," "," "," "," "," "," ",")"]
       | E_Zero      => []
       | E_Succ _    => ["("," ",")"]
       | E_Pred _    => ["("," ",")"]
       | E_IfZero _  => ["("," "," "," "," "," ",")"]
  in Term (t',fills') end

exception EMPTY_PATH
exception WAYWARD_PATH
fun insert _ [] _ = raise EMPTY_PATH (*inserts t' into t at position described by path*)
  | insert (Term (t,fill)) (pos::path) replacement =
      case t of 
        E_Var _                 => (case pos of
				      0 => replacement
				    | _ => raise WAYWARD_PATH
				   )
      | E_Lam (v,ty,t')         => (case pos of
				      0 => replacement
				    | 3 => (Term (E_Lam (v,ty,insert t' path replacement),fill))
				    | _ => raise WAYWARD_PATH
				   )
      | E_App (t1,t2)           => (case pos of
				      0 => replacement
				    | 1 => (Term (E_App (insert t1 path replacement,t2),fill))
				    | 2 => (Term (E_App (t1,insert t2 path replacement),fill))
				    | _ => raise WAYWARD_PATH
				   )
      | E_Un                    => (case pos of
				      0 => replacement
				    | _ => raise WAYWARD_PATH
				   )
      | E_Fix (v,ty,t')         => (case pos of
				      0 => replacement
				    | 3 => (Term (E_Fix (v,ty,insert t' path replacement),fill))
				    | _ => raise WAYWARD_PATH
				   )
      | E_InL (ty,t')           => (case pos of
				      0 => replacement
				    | 2 => (Term (E_InL (ty,insert t' path replacement),fill))
				    | _ => raise WAYWARD_PATH
				   )
      | E_InR (ty,t')           => (case pos of
				      0 => replacement
				    | 2 => (Term (E_InR (ty,insert t' path replacement),fill))
				    | _ => raise WAYWARD_PATH
				   )
      | E_Case (tL,vL,t',vR,tR) => (case pos of
				      0 => replacement
				    | 1 => (Term (E_Case (insert tL path replacement,vL,t',vR,tR),fill))
				    | 3 => (Term (E_Case (tL,vL,insert t' path replacement,vR,tR),fill))
				    | 5 => (Term (E_Case (tL,vL,t',vR,insert tR path replacement),fill))
				    | _ => raise WAYWARD_PATH
				   )
      | E_Zero                     => (case pos of
				         0 => replacement
				       | _ => raise WAYWARD_PATH
				      )
      | E_Succ t'                  => (case pos of
				         0 => replacement
				       | 1 => (Term (E_Succ (insert t' path replacement),fill))
				       | _ => raise WAYWARD_PATH
				      )
      | E_Pred t'                  => (case pos of
				         0 => replacement
				       | 1 => (Term (E_Pred (insert t' path replacement),fill))
				       | _ => raise WAYWARD_PATH
				      )
      | E_IfZero (t1,t2,t3)        => (case pos of
				         0 => replacement
				       | 1 => (Term (E_IfZero (insert t1 path replacement,t2,t3),fill))
				       | 2 => (Term (E_IfZero (t1,insert t2 path replacement,t3),fill))
				       | 3 => (Term (E_IfZero (t1,t2,insert t3 path replacement),fill))
				       | _ => raise WAYWARD_PATH
				      )

fun get_surr_ws (Term (_, fills)) =
  let 
    val left = hd fills
    val right = List.last fills
    fun whitespacePrefix' ans [] = ans
      | whitespacePrefix' ans (x :: xs) = 
          case x of
              #"(" => ans
            | #")" => ans
            (*since contents of fill are either brackets or whitespace*)
            | _    => whitespacePrefix' (x::ans) xs
    fun wsp f1 f2 =(*order matters since tab != space visually*)
      String.implode o f2 o (whitespacePrefix' []) o f1 o String.explode
    fun id x = x
    val left' = wsp id List.rev left
    val right' = wsp List.rev id right
  in (left', right') end

fun add_surr_ws (Term (t, fills)) (left', right') =
  let 
    val tailrev = List.rev o tl
    val midsection = tailrev o tailrev
    val left'' = String.concat [left', hd fills]
    val right'' = String.concat [List.last fills, right']
  in Term (t, left'' :: ((midsection fills) @ [right''])) end

end
