(*
#$Id: prefactor.sml 34 2011-07-21 17:52:28Z ns441 $
#
#Main file for the "prefactor" experiment. The parsing bits of this file are
#based on the ml-yacc tutorial.
#
#Nik Sultana
#www.cl.cam.ac.uk/~ns441
#
*)

structure Prefactor :
sig
  val parse_file: string -> PCFP1_Syntax.term
  val parse_exp: string -> PCF_Parser.result
  (*parses then pretty-prints a file*)
  val parse_then_print: string -> string
  (*"points" to a subterm*)
  val point: string -> PCFP1_Syntax.index -> PCFP1_Syntax.range -> unit
  val main: string * string list -> OS.Process.status
  val refactor: string -> PCFP1_Syntax.index -> PCFP1_Syntax.range -> string ->
    string -> PCFP1_Syntax.typ -> string PCFP1_Syntax.trm -> PCFP1_Syntax.term option
end =
struct

exception PCF_ERROR;

val VERSION =
  let
    val revision = "$Rev: 34 $"
  in String.substring (revision, 6, String.size revision - 8) end

(**
 * number of tokens to look-ahead for error-correction
 *)
val LOOKAHEAD = ref 0

local  
  fun printError file (msg, line, col) =
    print (file ^ "[" ^ Int.toString line ^ ":" ^
      Int.toString col ^ "] " ^ msg ^ "\n");
  fun parse file grab =
    PCF_Parser.parse
      (!LOOKAHEAD,
       PCF_Parser.makeLexer grab file,
       printError file,
       file)
       handle PCF_Parser.ParseError => raise PCF_ERROR;
in  
  (**
   * @param expression string encoding some PCFP1 expression
   * returns parsed contents of expression
   *)
  fun parse_exp expression =
    let  
      val current_pos = ref 0;
      fun get_next numchars =
        let
          val s = String.extract (expression, !current_pos, SOME numchars);
          val _ = current_pos := !current_pos + numchars;
        in s end
      val grab : int -> string = fn n =>
        let
          val remaining = String.size expression - !current_pos;
        in
          if !current_pos = n then
            ""
          else if remaining < n then
            get_next remaining
          else
            get_next n
        end
      val (tree, _) = parse "(EXPRESSION)" grab
    in tree end  

  (**
   * @param file filename being parsed; this is used in error messages
   * returns parsed contents of file
   *)
  fun parse_file file =
    let  
      val input_stream = TextIO.openIn file;
      val grab : int -> string = fn n =>
        if TextIO.endOfStream input_stream then
          ""
        else
          TextIO.inputN (input_stream, n);
      val (tree, _) = parse file grab
      val _ = TextIO.closeIn input_stream;
    in PCFP1_Syntax.term_of tree end
end (*end local*)

fun parse_then_print file = PCFP1_Syntax.print_trm (parse_file file)

fun point file i r =
  case PCFP1_Syntax.point (parse_file file) (i,r) of
     NONE => print "Could not locate (sub)term.\n"
   | SOME (term, path) =>
       print ("At path " ^ String.concat(map Int.toString path) ^
         " found subterm:\n[begins]\n" ^ PCFP1_Syntax.print_trm term ^
         "\n[ends]\n")

fun emptyCtxt _ = NONE

(*originally we have the unit program*)
val program : PCFP1_Syntax.term ref =
  ref (PCFP1_Syntax.Term (PCFP1_Syntax.E_Un,[]))

(**
 * returns "term option"; i.e. either no changes were made or the refactored
 * program is returned.
 *)
fun refactor file index range newvar1 newvar2 newtype newterm =
  let
    val _ = program := parse_file file
    val (term, path) = Option.valOf (PCFP1_Syntax.point (!program) (index,range))
    val typecheck =
      case CR.typeinfer (Biject.isabellise_term term) emptyCtxt of
          NONE =>
            (PCFP1_Syntax.debug_print "pointed-to term doesn't typecheck\n"
            ;false)
        | SOME _ =>
            (PCFP1_Syntax.debug_print "pointed-to term typechecks\n"
            ;true)
    val (changed, result) =
      if typecheck then
        let
          (*run the refactoring. prettify the output if the transformation was
          applied, otherwise leave the program unchanged.*)
          val refactored_exp =
            Biject.deisabellise_term (CR.r2 (Biject.isabellise_term term) newvar1
              newvar2 (Biject.isabellise_type newtype)
              (Biject.isabellise_term newterm) emptyCtxt)
          val changed = not (refactored_exp = term)
          (*insert whitespace in/around refactored expression*)
          val reformatted_refactored_exp =
            if changed then
              PCFP1_Syntax.add_surr_ws (PCFP1_Syntax.reformat refactored_exp)
                (PCFP1_Syntax.get_surr_ws term)
            else
              refactored_exp
           (*inject refactored expression back into the program*)
        in (changed, PCFP1_Syntax.insert (!program) path reformatted_refactored_exp) end
      else
        (false, !program)
  in
    (*the refactoring only returns the program if it was changed*)
    if changed then
      SOME result
    else
      NONE
  end

fun main (arg0, args) =
  let
	  val header = "prefactor (revision " ^ VERSION ^ ")\n\n"
	  val help_msg = "Try running:\n\t" ^ arg0 ^
	        " (operation) [arguments to the operation]\nFor the list of " ^
	        "operations supported and their arguments see the file USAGE that " ^
	        "is distributed together with prefactor.\n"
	  val args_errmsg = header ^ "Error: insufficient arguments.\n" ^ help_msg
	  fun error () = 
	    (print args_errmsg
	    ; OS.Process.failure)
	  fun operation () = hd args
	  fun file () = List.nth (args,1)
    fun index () = Option.valOf (Int.fromString (List.nth (args,2)))
    fun range () = Option.valOf (Int.fromString (List.nth (args,3)))
	in
    if List.null args then
       error ()
    else
      case (operation ()) of
          "parseprint" =>
            (print (parse_then_print (file ()))
            ; OS.Process.success)
        | "point" =>
            (point (file ()) (index ()) (range ())
            ; OS.Process.success)
        | "difftest" =>
            let
              val pp = parse_then_print (file ())
              (*-n asks 'echo' not to print an additional newline at the end of its output *)
              val cmd = "echo -n \""^pp^"\"| diff "^ file () ^" -"
              val _ = PCFP1_Syntax.debug_print cmd
            in OS.Process.system cmd end
        | "typeinfer" =>
            let
              val _ = program := parse_file (file ())
              val term = #1 (Option.valOf (PCFP1_Syntax.point
                              (!program) (index (), range ())))
            in
              case (CR.typeinfer (Biject.isabellise_term term) emptyCtxt) of
                  NONE =>
                   (print "Could not infer type.\n"
                   ; OS.Process.failure)
                | SOME ty =>
                   let
                    val ty' = Biject.deisabellise_type ty
                    val _ =
                      print ("Inferred type " ^ PCFP1_Syntax.print_typ
                        (PCFP1_Syntax.rebracket ty') ^ "\n")
                   in OS.Process.success end
            end
        | "refactor" =>
  	         let
               val newvar1 = List.nth (args, 4)
               val newvar2 = List.nth (args, 5)

               val newtype =
                 PCFP1_Syntax.type_of (parse_exp (List.nth (args, 6)))
               val newterm =
                 PCFP1_Syntax.term_of (parse_exp (List.nth (args, 7)))

               val outcome =
                 case refactor (file ()) (index ()) (range ()) newvar1 newvar2
                  newtype newterm of
                     NONE => OS.Process.failure
                   | SOME p =>
                       (program := p
                       ; OS.Process.success)
             in
               (print (PCFP1_Syntax.print_trm (!program))
               ; outcome)
             end
        | "refactor_patch" =>
             let
               val newvar1 = List.nth (args, 4)
               val newvar2 = List.nth (args, 5)

               val newtype =
                 PCFP1_Syntax.type_of (parse_exp (List.nth (args, 6)))
               val newterm =
                 PCFP1_Syntax.term_of (parse_exp (List.nth (args, 7)))

               val outcome =
                 case refactor (file ()) (index ()) (range ()) newvar1 newvar2
                  newtype newterm of
                     NONE => OS.Process.failure
                   | SOME p =>
                       (program := p
                       ; OS.Process.success)
               val cmd = "echo -n \"" ^ PCFP1_Syntax.print_trm (!program) ^
                            "\"| diff " ^ file () ^ " -"
			     in OS.Process.system(cmd)
			     end
		| _  => error ()
		end
end
